from invoke import task


@task
def upload_video(ctx):
    # Honestly not too sure this would be useful to have, as renders need to be local
    raise NotImplementedError()


@task
def make_playlist(ctx):
    # TODO use YouTube API Key to create playlists for a given semester
    raise NotImplementedError()



"""TODO docstring"""
from pathlib import Path

from ruamel.yaml import YAML

yaml = YAML()


class EditableFM:
    def __init__(self, base_path: Path, delim: str = "---"):
        self.base_path = Path(base_path)
        self.delim = delim

    def load(self, file: Path):
        self.fm = []
        self.content = []

        self.path = self.base_path / file

        file = self.path.open("r", encoding="utf-8").readlines()

        delims_seen = 0
        for line in file:
            if line.startswith(self.delim):
                delims_seen += 1
            else:
                if delims_seen < 2:
                    self.fm.append(line)
                else:
                    self.content.append(line)

        # Parse YAML, trying to preserve comments and whitespace
        self.fm = yaml.load("".join(self.fm))

    def dump(self):
        assert self.path, "You need to `.load()` first."

        with open(self.path, "w") as f:
            f.write(f"{self.delim}\n")
            yaml.dump(self.fm, f)
            f.write(f"{self.delim}\n")
            f.writelines(self.content)



import json
import os
import subprocess
from hashlib import sha256
from typing import Tuple
from pathlib import Path
from tempfile import TemporaryDirectory
import functools

import click
import requests
from click import Context

from seminar.utils import j2env, Config
from seminar.models import Meeting


shell = functools.partial(subprocess.call, shell=True, stdout=subprocess.DEVNULL)


@click.group(name="kaggle")
@click.pass_context
def cli(ctx: Context):
    r"""

    Parameters
    ----------
    ctx: :py:class:`click.Context`
        Click's built-in Context object. This is automatically passed by CLI execution.
    """
    tmp = ctx.obj.kaggle_dir = TemporaryDirectory()

    os.environ["KAGGLE_CONFIG_DIR"] = str(tmp.name)

    with (tmp / "kaggle.json").open("w", encoding="utf-8") as kaggle:
        config = json.loads(os.environ["KAGGLE_CONFIG"])
        json.dump(config, kaggle)

    os.chmod(tmp / "kaggle.json", "600")


@cli.command()
@click.pass_context
def kernel_metadata(ctx: Context, mtg: Meeting):
    r"""Generates `kernel-metadata.json` for each Kaggle competition.

    More info on the Kaggle API `kernel-metadata.json` file:
        https://github.com/Kaggle/kaggle-api/wiki/Kernel-Metadata

    Parameters
    ----------
    ctx: :py:class:`click.Context`
        Click's built-in Context object. This is automatically passed by :py:func:`ctx.invoke`.
    mtg: :py:class:`seminar.models.Meeting`
        The :py:class:`Meeting` to generate the Kaggle Kernel slug.
    """
    if "kaggle" not in mtg or mtg.kaggle is None:
        # TODO check for non-empty notebook
        return

    _username = ctx.obj.settings.kaggle.username

    metadata_j2 = j2env.get_template("kaggle/kernel-metadata.json.j2")

    if type(mtg.kaggle) == bool:
        mtg.kaggle = {}

    use_competitions = mtg.kaggle.get("competitions", False)

    mtg.kaggle["competitions"] = mtg.kaggle.get("competitions", [])
    mtg.kaggle["datasets"] = mtg.kaggle.get("datasets", [])
    mtg.kaggle["kernels"] = mtg.kaggle.get("kernels", [])
    mtg.kaggle["enable_gpu"] = mtg.kaggle.get("enable_gpu", False)

    if use_competitions:
        mtg.kaggle["competitions"].append(slug_competition(ctx.obj, mtg))

    kernel = mtg.as_file().with_name("kernel-metadata.json")
    with kernel.open("w", encoding="utf-8") as f:
        kwargs = {
            "username": _username,
            "slug": slug_kernel(ctx.obj, mtg),
            "kaggle": mtg.kaggle,
            "notebook": mtg.filename,
        }
        json.dump(json.loads(metadata_j2.render(**kwargs)), f, indent=2)


def _pull_kernel(ctx: Context, mtg: Meeting):
    r"""Pull Kaggle Kernel in `kernel-metadata.json` to be diff'd.

    Parameters
    ----------
    ctx: :py:class:`click.Context`
        Click's built-in Context object. This is automatically passed by :py:func:`ctx.invoke`.
    mtg: :py:class:`seminar.models.Meeting`
        The :py:class:`Meeting` to generate the Kaggle Kernel slug.
    """
    _username = ctx.obj.settings.kaggle.username

    existence_test = requests.get(
        f"https://kaggle.com/{_username}/{slug_kernel(ctx.obj, mtg)}"
    )

    if existence_test.status_code != requests.codes.OK:
        return None
    else:
        kaggle_cmd = "kaggle k pull -p /tmp/"
        shell(f"{kaggle_cmd} {_username}/{slug_kernel(ctx.obj, mtg)}")
        return Path("/tmp") / f"{slug_kernel(ctx.obj, mtg)}{mtg.workbook}"


def _diff_kernel(ctx: Context, mtg: Meeting, remote: Path) -> bool:
    r"""Uses `sha256` on local Workbook and Kaggle Kernel to determine if they differ.

    Parameters
    ----------
    ctx: :py:class:`click.Context`
        Click's built-in Context object. This is automatically passed by :py:func:`ctx.invoke`.
    mtg: :py:class:`seminar.models.Meeting`
        The :py:class:`Meeting` know which notebook to source when diff'ing notebooks.
    remote: :py:class:`pathlib.Path`
        The notebook currently hosted on Kaggle Kernels.

    Returns
    -------
    hash_equality: bool
        Returns ``True`` if the local and remote Workbooks differ.
    """
    local = ctx.obj.path / mtg.workbook

    remote_hash = sha256(open(remote, "rb").read()).hexdigest()
    local_hash = sha256(open(local, "rb").read()).hexdigest()

    return remote_hash != local_hash


@cli.command()
@click.pass_obj
def push_kernel(ctx: Context, mtg: Meeting):
    r"""Pushes a Meeting's local Workbook to Kaggle Kernels.

    Parameters
    ----------
    ctx: :py:class:`click.Context`
        Click's built-in Context object. This is automatically passed by :py:func:`ctx.invoke`.
    mtg: :py:class:`seminar.models.Meeting`
        The :py:class:`Meeting` to generate the Kaggle Kernel slug.
    """
    path = _pull_kernel(ctx, mtg)
    diff = _diff_kernel(ctx, mtg, path) if path else True

    if diff:
        cwd = Path().absolute()
        os.chdir(mtg.as_dir())
        shell("kaggle k push")
        os.chdir(cwd)
    else:
        raise ValueError("Kernels are the same")


@cli.command()
@click.pass_context
def create_competition(ctx: Context, mtg: Meeting) -> bool:
    r"""TODO: Create a Meeting's associated Kaggle InClass Competition.

    Parameters
    ----------
    ctx: :py:class:`click.Context`
        Click's built-in Context object. This is automatically passed by :py:func:`ctx.invoke`.
    mtg: :py:class:`seminar.models.Meeting`
        The :py:class:`Meeting` to generate the Kaggle Kernel slug.
    """
    # NOTE this might be doable using https://github.com/puppeteer/puppeteer,
    #   Selenium will definitely work, though
    raise NotImplementedError()


@cli.command()
@click.pass_context
def accept_competition(ctx: Context, mtg: Meeting) -> bool:
    r"""TODO: Accept a Meeting's associated Kaggle InClass Competition rules.

    Parameters
    ----------
    ctx: :py:class:`click.Context`
        Click's built-in Context object. This is automatically passed by :py:func:`ctx.invoke`.
    mtg: :py:class:`seminar.models.Meeting`
        The :py:class:`Meeting` to generate the Kaggle Kernel slug.
    """
    # NOTE this might be doable using https://github.com/puppeteer/puppeteer,
    #   Selenium will definitely work, though
    raise NotImplementedError()


def slug_kernel(cfg: Config, mtg: Meeting) -> str:
    r"""Generates Kaggle Kernel slugs of the form: `<group>-<semester>-<filename>`.
    e.g. If we consider the Spring 2020 "Building AI, the Human Way" lecture, the slug
    would be: `core-fa19-building-ai-the-human-way`.

    Parameters
    ----------
    ctx: :py:class:`click.Context`
        Click's built-in Context object. This is automatically passed by :py:func:`ctx.invoke`.
    mtg: :py:class:`seminar.models.Meeting`
        The :py:class:`Meeting` to generate the Kaggle Kernel slug.

    Return
    ------
    slug: str
        The Kaggle Kernel slug. Viewable at kaggle.com/<kaggle-username>/<slug>
    """
    return f"{cfg.group.name}-{cfg.semester}-{mtg.filename}"


def slug_competition(cfg: Config, mtg: Meeting) -> str:
    r"""Kaggle InClass competitions are listed under general competitions. So, we take
    each meeting's `slug_kernel` and prepend `ORG_NAME` to that – so for AI@UCF, it we
    prepend `ucfai`.

    Parameters
    ----------
    ctx: :py:class:`click.Context`
        Click's built-in Context object. This is automatically passed by :py:func:`ctx.invoke`.
    mtg: :py:class:`seminar.models.Meeting`
        The :py:class:`Meeting` to generate the Kaggle Kernel slug.

    Return
    ------
    slug: str
        The Kaggle InClass Competition slug. Viewable at kaggle.com/c/<slug>
    """
    return f"{cfg.settings.org_name}-{slug_kernel(cfg, mtg)}"



from invoke import task

# TODO install the SendGrid API
#      https://github.com/sendgrid/sendgrid-python


@task
def make(ctx):
    # TODO Collect syllabi entries to generate email(s)
    raise NotImplementedError()


@task
def send(ctx):
    # TODO Send email to departments / list
    raise NotImplementedError()



""" TODO docstring """


def by_semester(sem):
    assert len(sem) >= 4

    rank = int(1e6)
    semesters = {
        "fa": 2,
        "sp": 1,
        "su": 0,
    }
    name, year = sem[0:2], int(sem[2:4])
    return rank - (len(semesters) * year + semesters[name])


def by_roles(officers, role):
    """Takes in two: `<sem>-<group>-<role>`. Returns ordering by `<sem>` and `<role>`."""
    officers = [x.lower().replace(" ", "-") for x in officers]

    rank = by_semester(role)
    rank *= len(officers)
    for idx, officer in enumerate(officers[::-1]):
        if role.endswith(officer):
            rank -= len(officers) - idx
            break

    return rank



"""TODO docstring"""
from .group import Group
from .meeting import Meeting

__all__ = [
    # "Syllabus",
    "Group",
    "Meeting",
]


def _inline_list(ls):
    from ruamel.yaml.comments import CommentedSeq
    from ruamel.yaml.scalarstring import DoubleQuotedScalarString

    ls = CommentedSeq(ls)
    for idx, item in enumerate(ls):
        if type(item) == str:
            ls[idx] = DoubleQuotedScalarString(item)
    ls.fa.set_flow_style()
    return ls


def _transform(v):
    if type(v) == list:
        pass
        # v = _inline_list(v)
    elif type(v) == dict:
        for key in v.keys():
            v[key] = _transform(v[key])
    elif type(v) == str:
        from ruamel.yaml.scalarstring import DoubleQuotedScalarString

        v = DoubleQuotedScalarString(v)

    return v


def to_yaml(model, path):
    """Handles dumping `group.yml` and `syllabus.yml` to files."""
    import json

    from ruamel.yaml import YAML
    from ruamel.yaml.comments import CommentedMap
    from ruamel.yaml.representer import FoldedScalarString

    yaml = YAML()
    yaml.width = 78
    yaml.preserve_quotes = True

    if type(model) != list:
        model = [model]

    model_cm = []
    for idx, part in enumerate(model):
        part = CommentedMap(json.loads(part.json()))
        part.yaml_set_start_comment("Check the Docs for details on editing this!")

        try:
            abstract = part["abstract"] or "We're filling this out!"
            part["abstract"] = FoldedScalarString(abstract)
        except KeyError:
            pass

        for key in part.keys():
            part[key] = _transform(part[key])

        model_cm.append(part)

    if len(model) == 1 and not isinstance(model[0], Meeting):
        model_cm = model_cm[0]

    yaml.dump(model_cm, path.open("w", encoding="utf-8"))



import json

from jinja2 import Environment, PackageLoader

j2env = Environment(
    loader=PackageLoader("seminar", "templates"),
)
j2env.filters["jsonify"] = json.dumps


requests_headers = {
    "user-agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:15.0) Gecko/20100101 Firefox/15.0.1",
}


from .config import Config
from . import ucfcal, sort
from .editFM import EditableFM

__all__ = [
    "ucfcal",
    "EditableFM",
]



__version__ = "0.1.1"

import click

from seminar import models
from seminar import apis

from seminar.utils import Config

pass_config = click.make_pass_decorator(Config, ensure=True)

from seminar.commands import group, meeting, markdown, notebook, paper
from seminar import cli



"""Commands that perform operations over groups (analogous to courses).

Some commands strictly operate over the ``group.yml`` or ``syllabus.yml`` files,
while others enact individual commands over each meeting (analogous to a course
meeting date) item.

.. note:: Be sure to take a look at the general syntax of :ref:`seminar` to ensure you're
          able to interact with the appropriate groups/courses.

"""
import re
import shutil
from pathlib import Path

import click
from click import Context
import pandas as pd

from seminar import models
from seminar.utils import Config


@click.group(name="group")
@click.pass_context
def cli(ctx: Context):
    r"""Performs administrative tasks for a given semester in the Group."""
    pass


@cli.command()
@click.pass_obj
def validate(cfg: Config) -> None:
    r"""Validates the Group's Syllabus is typed properly, sorted, etc."""
    ondisk = list(filter(Path.is_dir, cfg.path.iterdir()))
    empty = len(ondisk) == 0

    group = cfg.group

    g_timedelta = pd.Timedelta(str(group.startdate.time()))
    g_date = group.startdate
    g_weekdelta = pd.Timedelta(7)

    prev_date = g_date
    for idx, meeting in enumerate(cfg.syllabus):
        if re.search(r"meeting\d\d", meeting.filename):
            continue
        if empty or meeting.room == "":
            meeting.room = ctx.obj.group.room
        if g_date.time().hour != 0 and meeting.date.time().hour == 0:
            meeting.date += g_timedelta
        if (meeting.date - prev_date) % g_weekdelta != pd.Timedelta(0):
            print(
                "Not a N-weekly gap. "
                f"The current gap for {meeting} is {(meeting.date - prev_date)}. "
                "Not a failure condition, but take note."
            )
        prev_date = meeting.date

    g_authors = set(map(str.lower, cfg.group.authors))
    for meeting in cfg.syllabus:
        if re.search(r"meeting\d\d", meeting.filename):
            continue
        m_authors = set(map(str.lower, meeting.authors))
        # TODO check how the authors are missing
        if not m_authors and meeting.title and meeting.filename:
            _msg = f"Meeting {meeting} has no authors."
        elif not m_authors.intersection(g_authors):
            _msg = f"Meeting {meeting} authors not in Group {ctx.obj.group} authors."
        else:
            continue
        print(_msg)

    try:
        n_mtgs = len(cfg.syllabus)
        n_uniq = len({m.date for m in cfg.syllabus})
        assert n_mtgs == n_uniq
    except AssertionError:
        pass

    ctx.obj.syllabus = sorted(cfg.syllabus, key=lambda m: m.date)

    # region Write out validated syllabus
    syllabus_yml = cfg.path / "syllabus.yml"
    models.to_yaml(cfg.syllabus, syllabus_yml)
    # endregion


@cli.command()
@click.pass_obj
def touch(cfg: Config):
    r"""Mimics Unix `touch` and creates/updates the Group."""
    from seminar.apis import hugo

    # region Make this semester's meeting directories
    # In the repository
    cfg.path.mkdir(exist_ok=True)
    # On the website
    hugo.touch_semester(ctx)
    # endregion

    # region Write to `group.yml`
    group_yml = cfg.path / "group.yml"
    models.to_yaml(cfg.group, group_yml)
    # endregion

    # region Write to `syllabus.yml`
    syllabus_yml = cfg.path / "syllabus.yml"
    models.to_yaml(cfg.syllabus, syllabus_yml)
    # endregion

    # region Create and/or rename the Meeting directories
    # NOTE this section strictly creates directories and renames. If you want to create
    #  actual meeting files, use `meeting.touch`.
    def getid(p):
        return str((p / ".metadata").open("r", encoding="utf-8").read())

    syllid = {str(m.id): cfg.path / str(m) for m in cfg.syllabus}
    ondisk = sorted(list(filter(Path.is_dir, cfg.path.glob("??-??-*/"))))
    ondisk = {getid(p): p for p in ondisk}

    for id, meeting in syllid.items():
        try:
            ondisk[id].rename(meeting)
        except (KeyError, FileNotFoundError):
            meeting.mkdir()
            (meeting / ".metadata").open("w", encoding="utf-8").write(str(id))
            ondisk[id] = meeting

        prv = ondisk[id].stem[6:]  # only consider filenames
        needs_rename = filter(lambda x: x.stem.startswith(prv), meeting.iterdir())

        new = meeting.stem[6:]
        for child in needs_rename:
            ext = "".join(child.suffixes)
            child.rename(child.parent / f"{new}{ext}")
    # endregion

    # region Update Author Profiles on the website
    for author in cfg.group.authors:
        hugo.touch_author(ctx, author.lower())
    # endregion


@cli.command()
@click.pass_context
def new_semester(ctx: Context) -> str:
    r"""Creates a new semester. Filters the {group}.yml context for semester creation. Calls `group.touch` post completion."""
    from seminar.models import Group, Meeting
    from seminar.utils.ucfcal import make_schedule

    cfg = ctx.obj

    assert not isinstance(cfg.group, Group), "Already made a new semester."

    kwargs = {"name": cfg.group, "semester": cfg.semester}
    group = Group(**kwargs)

    calendar = make_schedule(group, new_semester=True)
    syllabus = [
        Meeting(title=f"meeting{idx:02d}", filename=f"meeting{idx:02d}", date=date)
        for idx, date in enumerate(calendar)
    ]

    cfg.group = group
    cfg.semester = group.semester
    if cfg.path.parts.count(cfg.semester) < 1:
        cfg.path /= group.semester
    cfg.syllabus = syllabus

    ctx.invoke(touch)


# @task(pre=[preprocess], post=[status_dump])
# @click.command()
# def publish(ctx):
#     """Prepares "camera-ready" versions of the Group for public consumption."""
#     pass


@cli.command()
@click.pass_context
def cleanup(ctx: Context) -> str:
    r"""Cleans up excess files, roles, and the like for a Group."""
    from seminar.apis import hugo
    from seminar.commands import notebook, markdown

    # region Cleanup lingering syllabus entries and directories
    def isa_template(s: str):
        return re.search(r"meeting\d\d", s)

    ctx.obj.syllabus = list(
        filter(lambda m: not isa_template(m.filename), ctx.obj.syllabus)
    )

    dirs = filter(Path.is_dir, ctx.obj.path.iterdir())
    template_dirs = filter(lambda p: isa_template(p.stem), dirs)
    template_dirs = [shutil.rmtree(str(p)) for p in template_dirs]
    # endregion

    # region Cleanup unused default files (e.g. blank Notebooks, SimpleSummaries, etc.)
    for idx, meeting in enumerate(ctx.obj.syllabus):
        ctx.invoke(markdown.cleanup, meeting)
        ctx.invoke(notebook.cleanup, meeting)
        try:
            items = list(map(str, list(meeting.as_file().parent.iterdir())))
            assert len(items) > 0  # avoid repeating code
        except (AssertionError, FileNotFoundError) as e:
            pass
        finally:
            del ctx.obj.syllabus[idx]
            continue

        try:
            ends_markdown = map(lambda x: x.endswith("md"), items)
            ends_notebook = map(lambda x: x.endswith("ipynb"), items)
            assert any(ends_markdown) or any(ends_notebook)
        except AssertionError:
            shutil.rmtree(str(meeting.as_file().parent))
            del ctx.obj.syllabus[idx]
    # endregion

    ctx.invoke(hugo.cleanup_authors)

    # TODO rewrite cleaned syllabus/group
    ctx.invoke(touch)



"""TODO docstring."""
from typing import List
import string

import requests
from requests import codes as status_codes
from jinja2 import Template

from seminar.utils import requests_headers


def dashify(string: str) -> str:
    return "-".join(string.split("_"))


def downcase_dashify_name(name: str) -> str:
    # latter bit is to keep those symbols from being stripped
    for punc in set(string.punctuation) - {
        "-",
        "_",
    }:
        name = name.replace(punc, "")

    name = name.replace(" ", "-")
    name = name.replace("_", "-")

    # partially inspired by: https://stackoverflow.com/a/40382663
    name_len = len(name)

    def is_lower_around(n: int) -> bool:
        prev_char = name[n - 1].islower()
        next_char = (n + 1) < name_len and name[n + 1].islower()
        return prev_char or next_char

    start = 0
    parts = []
    for c in range(1, name_len):
        if name[c].isupper() and is_lower_around(c):
            parts.append(name[start:c])
            start = c
    parts.append(name[start:])

    name = "-".join(parts)

    return name.lower()


def test_url(slug: str, test_urls: List[str] = [""]) -> str:
    if not slug:
        return ""

    for test in test_urls:
        url = Template(test).render(slug=slug) if test != "" else slug
        query = requests.get(url, headers=requests_headers)
        if query.status_code == status_codes.OK:
            return query.url

    raise ValueError(f"Could not find {slug} at any test URLs: {test_urls}")



import os


ci_envs = [
    "CI",  # GitHub Actions, Travis CI, CircleCI, Cirrus CI, GitLab CI, AppVeyor, CodeShip, dsari
    "BUILD_NUMBER",  # Jenkins, TeamCity
    "RUN_ID",  # TaskCluster, dsari
]


def is_ci() -> bool:
    """Detects if we're in a CI environment or not.

    Based on: https://github.com/watson/ci-info/blob/074a5a9f2fae5f066abb6fb3c9e59992fee62c17/index.js#L52
    """
    return any([bool(os.environ.get(ci, False)) for ci in ci_envs])



"""TODO docstring"""
from hashlib import sha256
from typing import Union

import click
from click import Context

from seminar.apis import hugo
from seminar.models import Meeting
from seminar.utils import search, j2env, Config
from seminar import pass_config

from seminar.commands import query_args, query_kwargs


@click.group(name="markdown")
@click.option(*query_args, **query_kwargs)
@pass_config
def cli(cfg, query: str = ""):
    r"""Interact with a Meeting’s SummaryFile.

    Specify the name of a :class:`..models.Meeting` to perform operations over the
    SummaryFiles related to said :class:`..models.Meeting`.
    """
    search.for_meeting(cfg, query)
    pass


@cli.command()
@click.pass_context
def make_summaryfile(ctx: Context, mtg: Meeting):
    r"""Creates the default SummaryFile for a Meeting."""
    ext = ctx.obj.settings.suffixes.simplesummary
    path = mtg.as_file(prefix=ctx.obj.path).with_suffix(ext)

    if not path.exists() or path.stat().st_size == 0:
        simple_instructions = j2env.get_template("summaryfile.md.j2")
        with path.open("w") as f:
            f.write(simple_instructions.render(meeting=mtg))


@cli.command()
@click.pass_context
def make_post(ctx: Context, mtg: Meeting, weight=-1):
    r"""Prepares to export the Meeting SummaryFile to the web."""
    ext = ctx.obj.settings.suffixes.simplesummary
    path = mtg.as_file(prefix=ctx.obj.path).with_suffix(ext)

    try:
        with path.open("r", encoding="utf-8") as f:
            md = f.read()
    except FileNotFoundError:
        return False

    if not diff_summaryfile(ctx.obj, mtg):
        return False

    ctx.invoke(hugo.touch_meeting, mtg=mtg, body=md, weight=weight, from_nb=False)

    return True


def diff_summaryfile(cfg: Config, mtg: Meeting) -> None:
    r"""Determines if the Meeting's SummaryFile differs the default one."""
    tmpl = j2env.get_template("summaryfile.md.j2").render(meeting=mtg)

    try:
        ext = cfg.settings.suffixes.simplesummary
        path = mtg.as_file().with_suffix(ext)
        with path.open("r", encoding="utf-8") as f:
            disk = f.read()
    except FileNotFoundError:
        return True

    tmpl_sha = sha256(tmpl.encode("utf-8")).hexdigest()
    disk_sha = sha256(disk.encode("utf-8")).hexdigest()

    return tmpl_sha != disk_sha


@cli.command()
@click.pass_context
def cleanup(ctx: Context, mtg: Meeting) -> None:
    r"""Delete empty Markdown files associated with .Meeting."""
    if diff_summaryfile(ctx.obj, mtg):
        ext = ctx.obj.settings.suffixes.simplesummary
        mtg.as_file().with_suffix(ext).unlink(missing_ok=True)



"""Notebooks are the bread-and-butter of teaching folks to code. This command group
provides a way to generate Solutionbooks then render them as:

* Workbooks, to enable interactive tutorial/self-study sessions.
* Markdown Posts, to enable publication on the web.

"""
import json
from typing import Union
from pathlib import Path

import click
from click import Context
import nbformat as nbf
from jinja2 import Template
from nbconvert.exporters import MarkdownExporter, NotebookExporter
from nbconvert.preprocessors import TagRemovePreprocessor
from nbgrader.preprocessors import ClearOutput, ClearSolutions

from seminar.utils import j2env, search
from seminar.models import Meeting
from seminar.apis import hugo
from seminar.commands import query_args, query_kwargs


@click.group(name="notebook")
@click.option(*query_args, **query_kwargs)
@click.pass_context
def cli(ctx, query: str = "") -> None:
    r"""Interact with a Meeting's Jupyter Notebooks.

    Specify the name of a :class:`seminar.models.Meeting` to perform operations over the notebooks
    related to said :class:`seminar.models.Meeting`.
    """
    search.for_meeting(ctx.obj, query)
    return None


def _default_notebook(ctx: Context, mtg: Meeting) -> nbf.NotebookNode:
    r""""""
    nb = nbf.v4.new_notebook()

    # Inject Heading
    html_header = j2env.get_template("notebooks/header.html.j2")
    banner_url = Template(ctx.obj.settings.hugo.banner_url).render(
        group=ctx.obj.group, meeting=mtg
    )
    header = html_header.render(banner_url=banner_url, meeting=mtg, group=ctx.obj.group)
    header_metadata = {"title": mtg.title, "tags": ["nb-title", "template"]}
    nb.cells.insert(0, nbf.v4.new_markdown_cell(header, metadata=header_metadata))

    # Inject data-loading cell
    from seminar.apis import kaggle

    py_dataset_path = j2env.get_template("notebooks/data-pathing.py.j2")
    dataset = py_dataset_path.render(slug=kaggle.slug_competition(ctx.obj, mtg))
    dataset_metadata = {"language": "python", "tags": ["template"]}
    nb.cells.insert(1, nbf.v4.new_code_cell(dataset, metadata=dataset_metadata))

    # Inject Notebook Metadata
    nb_metadata = j2env.get_template("notebooks/nb-metadata.json.j2")
    metadata = nb_metadata.render(meeting=mtg, group=ctx.obj.group)
    nb.metadata.update(json.loads(metadata))

    return nb


def _read_solutionbook(mtg: Meeting, path: Path) -> nbf.NotebookNode:
    r""""""
    standard = NotebookExporter()
    standard.register_preprocessor(
        TagRemovePreprocessor(remove_cell_tags=["template"]), enabled=True
    )

    nb, _ = standard.from_filename(str(path))
    nb = nbf.reads(nb, as_version=4)

    return nb


@cli.command()
@click.pass_context
def make_solutionbook(ctx, mtg: Meeting):
    r"""Ensures that all Solutionbooks have accurate headings, pathing, and metadata.

    **Solutionbooks** are the notebooks that make up a fully-complete workshop/lesson.
    Participants should be able to proceed through this notebook with no errors and
    achieve the expected [code] outcomes.

    """
    base_nb = _default_notebook(ctx, mtg)

    path = ctx.obj.path / mtg.solnbook
    # If the notebook doesn't exist, or it's empty
    if not path.exists() or path.stat().st_size == 0:
        nbf.write(base_nb, open(path, "w"))

    read_nb = _read_solutionbook(mtg, path)
    read_nb.cells = base_nb.cells + read_nb.cells[len(base_nb.cells) :]
    read_nb.metadata.merge(base_nb.metadata)

    nbf.write(read_nb, path.open("w"))


@cli.command()
@click.pass_context
def make_workbook(ctx, mtg: Meeting) -> None:
    r"""Generates a Workbook from a Solutionbook.

    Workbooks are stripped down Solutionbooks that, namely:

    * Have no output cells.
    * Replace ``### BEGIN SOLUTION ... ### END SOLUTION`` blocks with
      ``raise NotImplementedError()`` snippets for viewers to practice on.

    This enables participants to follow along during live teaching sessions or to work
    through the notebooks on their own time without exposure to the answers.

    .. todo: Migrate back to ``enforce_metadata=True``
    """
    workbook = NotebookExporter()

    preprocessors = [ClearOutput(), ClearSolutions(enforce_metadata=False)]
    for preprocessor in preprocessors:
        workbook.register_preprocessor(preprocessor, enabled=True)
    # TODO migrate back to `enforce_metadata=True`

    # workbook.register_preprocessor(ValidateNBGrader(), enabled=True)
    # this is only useful if we can migrate back to `enforce_metadata=True`

    try:
        nb, _ = workbook.from_filename(str(mtg.solnbook))
        nb = nbf.reads(nb, as_version=4)

        nbf.write(nb, mtg.workbook.open("w", encoding="utf-8"))
    except Exception:
        raise Exception(f"Workbook export failed on `{mtg}`.")


@cli.command()
@click.pass_context
def make_post(ctx: Context, mtg: Meeting, weight: int = -1) -> None:
    """Preprocess a Solutionbook and prepare it to post on the web."""
    as_post = MarkdownExporter(
        config={
            "extra_loaders": [j2env.loader],
            "template_file": "notebooks/to-post.md.j2",
            "no_prompt": True,
        }
    )
    as_post.register_preprocessor(
        TagRemovePreprocessor(remove_cell_tags=["nb-title"], enabled=True)
    )

    # Default to `git`-based "Last modified..."
    # lastmod = pd.Timestamp(name.stat().st_mtime, unit="s")
    # setattr(m, "lastmod", lastmod)

    try:
        nb, _ = as_post.from_filename(str(ctx.obj.path / mtg.solnbook))
    except FileNotFoundError:
        return False

    if not diff_notebook(ctx, mtg):
        return False

    ctx.invoke(hugo.touch_meeting, mtg=mtg, body=nb, weight=weight, from_nb=True)

    return True


def diff_notebook(ctx: Context, mtg: Meeting) -> bool:
    tmpl = _default_notebook(ctx, mtg)

    standard = NotebookExporter()
    disk, _ = standard.from_filename(str(mtg.solnbook))
    disk = nbf.reads(disk, as_version=4)

    return tmpl != disk


@cli.command()
@click.pass_context
def kaggle_metadata(ctx: Context, mtg: Meeting):
    r"""TODO docstring"""
    if not diff_notebook(ctx, mtg):
        return

    from seminar.apis import kaggle

    ctx.invoke(kaggle.kernel_metadata, mtg=mtg)


@cli.command()
@click.pass_context
def push_kaggle(ctx: Context, mtg: Meeting):
    r"""TODO docstring"""
    if not diff_notebook(ctx, mtg):
        return

    from seminar.apis import kaggle

    ctx.invoke(kaggle.push_kernel, mtg=mtg)


@click.command()
@click.pass_context
def cleanup(ctx: Context, mtg: Meeting) -> None:
    r"""TODO docstring"""
    tmpl = _default_notebook(ctx, mtg)

    try:
        disk = _read_solutionbook(mtg, mtg.solnbook)
    except (FileNotFoundError, nbf.NBFormatError):
        return

    rm = False
    rm = len(disk.cells) == 0 or tmpl.cells == disk.cells

    if rm:
        mtg.solnbook.unlink()
        mtg.workbook.unlink(missing_ok=True)
        mtg.kernel.unlink(missing_ok=True)



"""TODO doctstring."""
from typing import List
from pathlib import Path

import requests
import pandas as pd
from pydantic import (
    BaseModel,
    Field,
    root_validator,
)

from seminar.models import helpers


def _load_config() -> dict:
    from ruamel.yaml import YAML

    yaml = YAML()
    config = Path(__file__).parent.parent.parent / "config.yml"
    config = dict(yaml.load(config.open("r", encoding="utf-8")))
    return config


def infer_current_semester() -> str:
    r"""Infers the current semester based on the academic calendar.

    .. warning:: Currently, this is overfit to UCF, but it would be ideal to allow this
                 to be configurable in some reasonably easy manner.

    Returns
    -------
    semester: str
        The current semester in a `ttYY` where ``t`` is a 2-letter abbreviation of the
        current term and ``YY`` are the last two digits of the year.
    """
    config = _load_config()

    calendar_url = config["calendar"]["url"]
    url = requests.get(f"http://{calendar_url}/").url

    year, semester = url.split(f"{calendar_url}/")[-1].split("/")

    year = str(year)[-2:]
    sem = config["semesters"][semester]["short"]

    return f"{sem}{year}"


def infer_next_semester() -> str:
    r"""Infers the next semester based on today's date.

    .. warning:: Currently, this is overfit to UCF, but it would be ideal to allow this
                 to be configurable in some reasonably easy manner.

    Returns
    -------
    semester: str
        The current semester in a `ttYY` where ``t`` is a 2-letter abbreviation of the
        current term and ``YY`` are the last two digits of the year.
    """
    semester = infer_current_semester()

    semester, year = semester[:2], semester[-2:]

    config = _load_config()
    semesters = config["semesters"]

    semester = next(filter(lambda x: x.startswith(semester), semesters.keys()))

    if semester == "fall":
        year = f"{int(year) + 1}"

    sem = semesters[semester]["next"]
    sem = semesters[sem]["short"]

    return f"{sem}{year}"


def infer_startdate(semester: str = infer_next_semester()) -> pd.Timestamp:
    r"""Infer the startdate based on the semester's start.

    .. warning:: Currently, this is overfit to AI@UCF, but it would be ideal to allow
                 this to be configurable in some reasonably easy way.

    Parameters
    ----------
    semester: str
        **Optional.** Used to extract the start date based on the calendar.

    Returns
    -------
    startdate: :class:`pandas.Timestamp`
        Returns the anticipated start date of Meetings.
    """
    from seminar.utils.ucfcal import parse_calendar

    calendar, _ = parse_calendar(semester)
    return calendar[0] + pd.Timedelta(days=7)


class Group(BaseModel):
    r"""Groups are analogous to clubs and courses.

    This structure enables a single organization (or an instructor) to conduct multiple
    seminars in a coherent way that doesn't require lots of micromanaging of execution
    location, website management, and the like.

    Parameters
    ----------
    """
    name: str
    """The name of the Group."""
    room: str = ""
    """Meeting location."""

    semester: str = Field(default_factory=infer_next_semester)
    """Semester this group meets in."""
    startdate: pd.Timestamp = Field(default_factory=infer_startdate)
    """Start date, used to compute meetings based on pattern."""

    directors: List[str] = []
    """The Group leader(s) (or Lead Instructor(s))."""
    coordinators: List[str] = []
    """Analogous to Teaching Asistants."""
    guests: List[str] = []
    """**Optional.** Invited guest speakers."""
    advisors: List[str] = []
    """**Optional.** Experienced individuals who are advising the direction of this Seminar."""

    abstract: str = ""
    """A paragraph summary of the Group."""

    class Config:
        alias_generator = helpers.dashify

    @root_validator
    def _check_semester_and_startdate_match(cls, values):
        semester, startdate = values.get("semester"), values.get("startdate")
        if semester is None or startdate is None:
            raise ValueError(
                f"`semester` and `startdate` must be set. Got {(semester, startdate)}"
            )
        expected_startdate = infer_startdate(semester)
        exp_yr, exp_mo, _ = str(expected_startdate.date()).split("-")

        cur_yr, cur_mo, _ = str(startdate.date()).split("-")
        cur_hr, cur_mn, _ = map(int, str(startdate.time()).split(":"))

        # Check that we're setting dates that make sense for the semester; but also
        #   avoid overwriting manually set dates by checking the time.
        if (cur_hr == 0) and (exp_yr != cur_yr or exp_mo != cur_mo):
            values["startdate"] = expected_startdate

        return values

    def __str__(self) -> str:
        return self.name

    @property
    def authors(self) -> set:
        """A synthesized list of individuals who considered "eligible" to lead Meetings.
        These individuals must be present in one of the attributes: :attr:`directors`,
        :attr:`coordinators`, :attr:`guests`, or :attr:`advisors`.
        """
        return set(self.directors + self.coordinators + self.guests + self.advisors)



""" TODO docstring """
from typing import Union

import requests
import click

from seminar.models import Meeting
from seminar.utils import search, requests_headers
from seminar import pass_config
from seminar.commands import query_args, query_kwargs
from seminar.utils import Config


@click.group(name="paper", help="Strictly interacts with a Meeting's papers.")
@click.option(*query_args, **query_kwargs)
@click.pass_obj
def cli(cfg: Config, query: str = ""):
    r"""Interact with a Meeting’s assocaited readings.

    Specify the name of a :class:`seminar.models.Meeting` to perform operations over the
    readings related to said :class:`seminar.models.Meeting`.
    """
    search.for_meeting(cfg, query)


@cli.command()
@click.pass_obj
def download(cfg: Config, mtg: Meeting):
    """Downloads papers and names them based on keys for meetings."""
    if mtg.papers is None:
        return

    folder = cfg.path / mtg.path

    for title, link in mtg.papers.items():
        try:
            with (folder / f"{title}.pdf").open("wb") as pdf:
                pdf.write(requests.get(link, headers=requests_headers).content)
            # TODO Check for corruptness / PDF headers
        except ConnectionError:
            raise ConnectionError(
                f"Failed to connect to {link}, please check it produces a PDF."
            )



"""TODO docstring"""
import re
import shutil
from functools import partial
from pathlib import Path
from typing import Dict

import docker
import click
from click import Context
from docker.models.containers import ContainerCollection
from jinja2 import Template
from ruamel.yaml import YAML

from seminar.models import Meeting
from seminar.utils import EditableFM, sort, requests_headers


command = Template("hugo new --kind {{ kind }} {{ path }}")


def via_docker(func):
    def hugo_via_container(ctx: Context, *args, **kwargs):
        client = docker.from_env()

        import requests
        from requests.exceptions import RequestException

        s = ctx.obj.settings
        try:
            raise RequestException
            compose_yml = requests.get(
                f"{s.vcs.owner}/{s.web.repo}/raw/{s.web.branch}/docker-compose.yml",
                headers=request_headers,
            )
            compose_yml = YAML().load(compose_yml.content)
            image = compose_yml["services"]["hugo"]["image"]
        except (RequestException, KeyError):
            image = s.web.image

        client.images.pull(image.split(":")[0], image.split(":")[1])
        container = client.containers.create(
            image=image,
            auto_remove=False,
            detach=False,
            volumes={
                f"{s.web.src}": {
                    "bind": "/src",
                    "mode": "rw",
                },
            },
            working_dir="/src",
        )
        container.start()

        retval = func(ctx, *args, container=container, **kwargs)

        container.stop()
        container.remove()

        return retval

    hugo_via_container.__doc__ = func.__doc__

    return hugo_via_container


@via_docker
def touch_group(ctx: Context, container: ContainerCollection = None) -> EditableFM:
    r"""Touches a group's aggregation page.

    Parameters
    ----------
    ctx: :py:class:`click.Context`
        Click's built-in Context object. This is automatically passed by :py:func:`ctx.invoke`.
    container: :py:class:`docker.models.collections.ContainerCollection`
        Pass a running Docker container, as spawned by the ``docker`` Python library.


    Returns
    -------
    editableFM: EditableFM
        A Python-editble rendition of a Hugo post.
    """
    assert container
    group_path = f"groups/{ctx.obj.group.name}"
    new_group = command.render(
        kind="semester",
        path=group_path,
    )

    res = container.exec_run(new_author)
    output = res.output.decode("utf-8")

    # editor = EditableFM(f"{src}/content/{group_path}")
    # editor.load("_index.md")

    # NOTE Any programmatic edits to the Group should be made here

    # editor.dump()

    # return editor


@via_docker
def touch_semester(ctx: Context, container: ContainerCollection = None) -> None:
    r"""Creates a group's semester aggregation page.

    Parameters
    ----------
    ctx: :py:class:`click.Context`
        Click's built-in Context object. This is automatically passed by :py:func:`ctx.invoke`.
    container: :py:class:`docker.models.collections.ContainerCollection`
        Pass a running Docker container, as spawned by the ``docker`` Python library.


    Returns
    -------
    editableFM: EditableFM
        A Python-editble rendition of a Hugo post.
    """
    assert container

    group_path = f"groups/{ctx.obj.group.name}/{ctx.obj.group.semester}"
    new_semester = command.render(
        kind="semester",
        path=group_path,
    )

    expanded_path = f"{ctx.obj.settings.web.src}/content/{group_path}"
    Path(expanded_path).mkdir(exist_ok=True)

    res = container.exec_run(new_semester)
    output = res.output.decode("utf-8")

    editor = EditableFM(expanded_path)
    editor.load("_index.md")

    editor.fm["date"] = str(ctx.obj.group.startdate)
    editor.fm["location"] = ctx.obj.group.room
    editor.fm["abstract"] = ctx.obj.group.abstract

    editor.dump()


@via_docker
def touch_author(
    ctx: Context, author: str = "", container: ContainerCollection = None
) -> None:
    r"""Creates an author page everyone that contributes to a semester's content.

    Parameters
    ----------
    ctx: :py:class:`click.Context`
        Click's built-in Context object. This is automatically passed by :py:func:`ctx.invoke`.
    author: str
        The GitHub username of the author.
    container: :py:class:`docker.models.collections.ContainerCollection`
        Pass a running Docker container, as spawned by the ``docker`` Python library.


    Returns
    -------
    editableFM: EditableFM
        A Python-editble rendition of a Hugo post.
    """
    assert container

    author_path = f"authors/{author}/"
    new_author = command.render(
        kind="author",
        path=author_path,
    )

    res = container.exec_run(new_author)
    output = res.output.decode("utf-8")

    editor = EditableFM(f"{ctx.obj.settings.web.src}/content/{author_path}")
    editor.load("_index.md")

    groups = []
    for role in ctx.obj.settings.web.config.bot_taggable_roles:
        if author in map(str.lower, getattr(ctx.obj.group, f"{role.lower()}s")):
            groups = list(map(Template, ctx.obj.settings.web.roles))
            groups = [g.render(group=ctx.obj.group, role=role) for g in groups]
            break

    try:
        roles = [groups[-1]]
    except IndexError:
        roles = []

    editor.fm = _sort_and_update_metadata(
        ctx,
        editor.fm,
        editor.fm["user_groups"] + groups,
        editor.fm["teams"],
        editor.fm["roles"] + roles,
    )

    editor.dump()


@via_docker
def touch_post(
    ctx: Context,
    mtg: Meeting,
    weight: int = -1,
    container: ContainerCollection = None,
) -> EditableFM:
    r"""Prepares a Markdown post for the web. Body content is specified by the caller.

    Parameters
    ----------
    ctx: :py:class:`click.Context`
        Click's built-in Context object. This is automatically passed by :py:func:`ctx.invoke`.
    mtg: :py:class:`seminar.models.Meeting`
        The :py:class:`Meeting` to export to a post for the website.
    weight: int
        Hugo requires a weight to determine orderings (dates don't seem to suffice).
    container: :py:class:`docker.models.collections.ContainerCollection`
        Pass a running Docker container, as spawned by the ``docker`` Python library.

    Returns
    -------
    editableFM: :py:class:`seminar.utils.EditableFM`
        A Python-editble rendition of a Hugo post.
    """
    assert container

    meeting_path = f"groups/{ctx.obj.group.name}/{ctx.obj.semester}"
    meeting_file = f"{mtg.filename}.md"
    new_meeting_post = command.render(
        kind="group-meeting", path=f"{meeting_path}/{meeting_file}"
    )
    res = container.exec_run(new_meeting_post)
    output = res.output.decode("utf-8")

    editor = EditableFM(ctx.obj.settings.web.src / "content" / meeting_path)
    editor.load(meeting_file)

    return editor


@click.pass_context
def touch_meeting(
    ctx: Context, mtg: Meeting, body: str, weight: int, from_nb: bool = False
) -> None:
    editor = touch_post(ctx, mtg)

    editor.fm["title"] = mtg.title
    editor.fm["linktitle"] = mtg.title

    editor.fm["date"] = mtg.date.isoformat()
    # ctx.settings.hugo ordering prefers 1-based indexing
    if weight > -1:
        editor.fm["weight"] = weight + 1

    editor.fm["authors"] = mtg.authors

    editor.fm["urls"] = {}

    try:
        editor.fm["urls"]["video"] = mtg.urls.video
    except AttributeError:
        pass

    try:
        editor.fm["urls"]["slides"] = mtg.urls.slides
    except AttributeError:
        pass

    if from_nb:
        from seminar.apis import kaggle

        editor.fm["urls"]["github"] = str(mtg.solnbook)
        editor.fm["urls"]["kaggle"] = str(kaggle.slug_kernel(ctx.obj, mtg))
        editor.fm["urls"]["colab"] = str(mtg.workbook)

    editor.fm["location"] = mtg.room
    # editor.fm["cover"] = m.cover_image

    editor.fm["tags"] = mtg.tags
    editor.fm["abstract"] = mtg.abstract

    try:
        editor.fm["papers"] = {
            key: str(mtg.path / f"{key}.pdf") for key in mtg.papers.keys()
        }
    except (AttributeError):
        pass

    try:
        editor.content = ["\n", body]
    except NameError:
        pass

    editor.dump()


@click.pass_context
def cleanup_authors(ctx: Context) -> None:
    r"""Ensures that authors are properly credited and assocaited with correct semesters.

    Parameters
    ----------
    ctx: :py:class:`click.Context`
        Click's built-in Context object. This is automatically passed by :py:func:`ctx.invoke`.

    Returns
    -------
    nil: None
    """
    authors = Path(ctx.obj.settings.web.src) / "content/authors"

    marked_for_deletion = []
    for author in filter(Path.is_dir, authors.iterdir()):
        # import pdb; pdb.set_trace()
        editor = EditableFM(author)
        editor.load("_index.md")

        author = author.stem
        try:
            author_groups = set(editor.fm["user_groups"])
            author_roles = set(editor.fm["roles"])
            author_teams = set(editor.fm["teams"])
        except (TypeError, AssertionError) as e:
            marked_for_deletion += [author]

        for role in ctx.obj.settings.web.config.bot_taggable_roles:
            role_group = map(str.lower, getattr(ctx.obj.group, f"{role.lower()}s"))
            if author not in list(role_group):
                roles = [Template(r) for r in ctx.obj.settings.web.roles]
                roles = [r.render(group=ctx.obj.group, role=role) for r in roles]
                roles = set(filter(lambda x: role.lower() in x, roles))
                author_groups -= roles

        rm_roles = []
        for role in author_roles:
            has_role = list(filter(lambda x: x.endswith(role.lower()), author_groups))
            if not has_role:
                rm_roles.append(role)

        author_roles -= set(rm_roles)

        rm_teams = []
        for team in author_teams:
            on_team = list(filter(lambda x: x.startswith(team.lower()), author_groups))
            if not on_team:
                rm_teams.append(team)

        author_roles -= set(rm_teams)

        editor.fm = _sort_and_update_metadata(
            ctx, editor.fm, author_groups, author_teams, author_roles
        )

        editor.dump()

    for author in marked_for_deletion:
        shutil.rmtree(str(authors / author))


def _sort_and_update_metadata(
    ctx: Context, fm: dict, user_groups: list, teams: list, roles: list
) -> Dict:
    r"""Sorts and updates the groups, teams, and roles of authors.

    Parameters
    ----------

    Returns
    -------

    """
    # region Semi-Complex Group filtering/sorting
    groups = set(user_groups)

    sorter = partial(sort.by_roles, ctx.obj.settings.web.config.officers)

    non_sem = filter(lambda x: not re.match(r"(fa|sp|su)\d{2}", x), groups)
    non_sem_groups = sorted(list(non_sem))

    isa_sem = filter(lambda x: bool(re.match(r"(fa|sp|su)\d{2}", x)), groups)
    isa_sem_groups = sorted(list(isa_sem), key=sorter, reverse=True)

    fm["user_groups"] = non_sem_groups + isa_sem_groups
    # endregion

    # region Sort "teams" (semesterly activity)
    teams = list(set(teams))
    teams = sorted(teams, key=sort.by_semester, reverse=True)
    fm["teams"] = teams
    # endregion

    # region Sort "roles" alphabetically
    roles = sorted(list(set(roles)))
    fm["roles"] = roles
    # endregion

    return fm



"""The entrypoint of the `seminar` library. Handles general configuration and routing.

.. moduleauthor:: John Muchovej <jmuchovej@gmail.com>
"""
import os
from pathlib import Path

import click
from click import Context

from seminar.models import Group, Meeting
from seminar import group, meeting
from seminar.utils import environment
from seminar.utils.config import Config

from seminar.commands import group, meeting, notebook, markdown, paper
from seminar.models.group import infer_current_semester

APP_DIR = Path(click.get_app_dir("seminar-toolkit"))


def _workdir_update(cwd, update) -> Path:
    dirs = list(filter(Path.is_dir, cwd.iterdir()))
    if (cwd / update) in dirs:
        cwd /= update
    return cwd


def _check_or_set_semester(ctx, param, value):
    if value is None or len(value) == 0:
        return infer_current_semester()
    elif len(value) != 4:
        return click.BadParameter("Must be a 4-character representation.")
    elif not (value[:2] in ["sp", "su", "fa"] and value[2:].isdigit()):
        return click.BadParameter("Must match the following format: (sp|su|fa)XX.")
    return value


@click.group(
    invoke_without_command=True,
    context_settings={
        "help_option_names": ["-h", "--help"],
        # "allow_interspersed_args": True,
    },
)
@click.option(
    "-o",
    "--org",
    "organization",
    type=str,
    required=False,
    default="",
    help="Specifying the organization name helps locate the appropriate `config.yml`.",
)
@click.option(
    "-c",
    "--config",
    "config",
    type=Path,
    default=APP_DIR / "config.yml",
    help="The path to your desired `config.yml`.",
)
@click.option(
    "-g",
    "--group",
    "group",
    type=str,
    required=True,
    help="The group you're modifying.",
)
@click.option(
    "-s",
    "--semester",
    "semester",
    callback=_check_or_set_semester,
    type=str,
    required=True,
    default=infer_current_semester(),
    help="The semester you're modifying.",
)
@click.pass_context
@click.version_option()
@click.help_option()
def seminar(
    ctx: Context,
    organization: str,
    config: Path,
    group: str,
    semester: str,
) -> None:
    from ruamel.yaml import YAML

    yaml = YAML()

    ctx.obj = Config(ci=environment.is_ci())

    cwd = Path().absolute()
    if ctx.obj.ci:
        cwd /= semester
    else:
        cwd = _workdir_update(cwd, group)
        cwd = _workdir_update(cwd, semester)

    ctx.obj.path = cwd
    try_specific_app_dir = bool(organization) and APP_DIR.stem in str(config)
    if try_specific_app_dir and organization not in config.name:
        config = APP_DIR / organization / "config.yml"

    ctx.obj.settings = dict(yaml.load(open(config, "r")))

    os.chdir(cwd)

    try:
        group = ctx.obj.path / "group.yml"
        group = yaml.load(group.open("r", encoding="utf-8"))
        group = Group(**dict(group))
    except FileNotFoundError:
        group = ctx.params["group"]
        semester = ctx.params["semester"]
    finally:
        ctx.obj.group = group
        ctx.obj.semester = semester

    try:
        syllabus = ctx.obj.path / "syllabus.yml"
        syllabus = yaml.load(syllabus.open("r", encoding="utf-8"))
        syllabus = list(map(dict, syllabus))
        syllabus = [Meeting(**kwargs) for kwargs in syllabus]
    except FileNotFoundError:
        syllabus = []
    finally:
        ctx.obj.syllabus = syllabus

    if ctx.obj.ci:
        site_src = Path("/")
    else:
        site_src = ctx.obj.path.parent
        if semester in ctx.obj.path.stem:
            site_src = site_src.parent

    site_src /= ctx.obj.settings.web.repo

    ctx.obj.settings.web.src = site_src
    assert site_src.exists(), site_src

    ctx.obj.settings.web.config = site_src / ctx.obj.settings.web.config
    ctx.obj.settings.web.config = yaml.load(
        ctx.obj.settings.web.config.open("r", encoding="utf-8")
    )


seminar.add_command(group.cli)
seminar.add_command(meeting.cli)
seminar.add_command(notebook.cli)
seminar.add_command(markdown.cli)
seminar.add_command(paper.cli)



""" TODO docstring """
from pathlib import Path

import pandas as pd
import requests

from ruamel.yaml import YAML

_yaml = YAML()
_config = Path(__file__).parent.parent.parent / "config.yml"
_config = _yaml.load(_config.open("r", encoding="utf-8"))
del _yaml


def make_schedule(group, delta: int = 7, new_semester: bool = False):
    if not new_semester and group.startdate.time().hour == 0:
        raise ValueError("Please set a time for the meetings.")

    calendar, holidays = parse_calendar(group.semester)
    dates = [group.startdate]
    while dates[-1] < calendar.iloc[-1]:
        dates.append(dates[-1] + pd.Timedelta(days=delta))

    # Removes Holidays, if the array is non-empty.
    holidays = list(map(lambda x: x.date(), holidays))
    meeting_dates = list(filter(lambda x: x.date() not in holidays, dates))

    schedule = [pd.Timestamp(x) for x in meeting_dates]

    return schedule


def parse_calendar(semester: str) -> tuple:
    from jinja2 import Template

    calendar_url = Template(_config["calendar"]["format"])
    calendar_base_url = _config["calendar"]["url"]

    semesters = _config["semesters"]

    lookup = {v["short"]: k for k, v in semesters.items()}
    lookup_sm = lookup[semester[:2]]
    lookup_yr = 2000 + int(semester[-2:])
    semester = semesters[lookup_sm]

    # This is the URL for the calendar's JSON-based API. This will vary by institution.
    calendar_url = "http://" + calendar_url.render(
        url=calendar_base_url, year=lookup_yr, longname=lookup_sm
    )
    url = str(requests.get(calendar_url).url)
    url = url.replace(calendar_base_url, f"{calendar_base_url}/json")

    # UCF's JSON object holds all the "events" in an "events" identifier
    ucf_parsed = requests.get(url).json()["terms"][0]["events"]
    df_calendar = pd.DataFrame.from_dict(ucf_parsed)

    # The "summary" column of the new DataFrame holds the names of the events
    summary_mask = df_calendar["summary"]
    # Events are stored as arrays. It's easier to parse this way than reducing
    #   the list-containing column to something that isn't.
    starts = df_calendar.loc[summary_mask.str.contains("Classes Begin")].iloc[0]
    ends = df_calendar.loc[summary_mask.str.contains("Classes End")].iloc[0]

    # generate a DataFrame with all possible dates (to act like a calendar)
    calendar = pd.Series(
        pd.date_range(start=starts["dtstart"][:-1], end=ends["dtstart"][:-1])
    )

    # Remove the Holidays, since we assume students won't meet on those days.
    holidays = []
    for holiday in semester["holidays"]:
        lookup = df_calendar.loc[summary_mask.str.contains(holiday)]
        for _, day2rm in lookup.iterrows():
            beg = day2rm["dtstart"][:-1]
            end = day2rm["dtend"][:-1] if day2rm["dtend"] else beg
            holidays.append(pd.Series(pd.date_range(start=beg, end=end)))

    holidays = pd.concat(holidays).reset_index(drop=True)

    return calendar, holidays



"""TODO docstring"""
import re

import click
from click import Context

from seminar import pass_config
from seminar.utils import search
from seminar.commands import query_args, query_kwargs


@click.group(name="meeting")
@click.option(*query_args, **query_kwargs)
@click.pass_context
def cli(ctx, query: str = ""):
    r"""Runs the Meeting pipeline for whichever meeting(s) is(are) requested."""
    search.for_meeting(ctx.obj, query)
    pass


@cli.command()
@click.pass_context
def touch(ctx: Context):
    r"""Mimics Unix `touch` and creates/updates Meetings.

    Generates and standardizes Solutionbooks, SummaryFiles, and auxiliary
    :py:class:`Meeting` contents.
    """
    from seminar.commands import notebook, markdown, paper

    for meeting in ctx.obj.meetings:
        # ctx.params["query"] = meeting

        # Primary Meeting contents, strictly one of these should be filled
        ctx.invoke(notebook.make_solutionbook, mtg=meeting)
        ctx.invoke(markdown.make_summaryfile, mtg=meeting)

        # Auxiliary Meeting contents, none of these must to be used
        ctx.invoke(notebook.kaggle_metadata, mtg=meeting)
        ctx.invoke(paper.download, mtg=meeting)


@cli.command()
@click.pass_context
def publish(ctx: Context):
    r"""Prepares "camera-ready" versions of Meetings for public consumption.

    Compiles a [semi-]completed Solutionbook or SummaryFile into formats ready for
    publication on the web and other useful platforms. (e.g. if using Solutionbooks,
    an advantageous publication venue is Kaggle.)

    """
    from seminar.commands import notebook, markdown
    from seminar.apis import hugo

    start = -1 if len(ctx.obj.meetings) < len(ctx.obj.syllabus) else 0

    for weight, meeting in enumerate(ctx.obj.meetings, start=start):
        if re.match("meeting\d\d", meeting.filename):
            # TODO provide failure message
            continue

        ctx.invoke(notebook.make_workbook, mtg=meeting)
        ctx.invoke(notebook.push_kaggle, mtg=meeting)

        export_nb = ctx.invoke(notebook.make_post, mtg=meeting, weight=weight)
        export_md = ctx.invoke(markdown.make_post, mtg=meeting, weight=weight)

        # TODO check XOR(export_nb, export_md)
        try:
            if bool(export_nb) == bool(export_md):
                # TODO write blank to disk
                ctx.invoke(
                    hugo.touch_meeting,
                    mtg=meeting,
                    body="",
                    weight=weight,
                    from_nb=False,
                )
                raise ValueError(
                    "Use of Solutionbook OR Markdown is mutually exclusive. Writing only meeting metadata."
                )
        except ValueError as e:
            print(str(e))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['seminar',
 'seminar.apis',
 'seminar.commands',
 'seminar.models',
 'seminar.utils']

package_data = \
{'': ['*'],
 'seminar': ['templates/*',
             'templates/group/*',
             'templates/kaggle/*',
             'templates/notebooks/*']}

install_requires = \
['Jinja2>=2.11.2,<3.0.0',
 'Pygments>=2.7.3,<3.0.0',
 'click>=7.1.2,<8.0.0',
 'docker>=4.4.1,<5.0.0',
 'emoji>=0.6.0,<0.7.0',
 'jupyter>=1.0.0,<2.0.0',
 'kaggle>=1.5.10,<2.0.0',
 'nbconvert>=6.0.7,<7.0.0',
 'nbformat>=5.0.8,<6.0.0',
 'nbgrader>=0.6.1,<0.7.0',
 'numpy>=1.19.5,<2.0.0',
 'pandas>=1.2.0,<2.0.0',
 'pandoc>=1.0.2,<2.0.0',
 'pydantic>=1.7.3,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'ruamel.yaml>=0.16.12,<0.17.0',
 'termcolor>=1.1.0,<2.0.0']

entry_points = \
{'console_scripts': ['seminar = seminar.cli:seminar']}

setup_kwargs = {
    'name': 'seminar-toolkit',
    'version': '0.1.1',
    'description': 'Optimized to manage Jupyter Notebook driven courseware with SSGs.',
    'long_description': '',
    'author': 'John Muchovej',
    'author_email': '5000729+jmuchovej@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://john.muchovej.com/projects/seminar/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



"""A recursive dictionary generator which allows for accessing keys via "."s .

Implementation mostly sourced from: https://stackoverflow.com/a/32107024/2714651.
"""


class Config(dict):
    """A recursive dictionary generater that allows for accessing keys via dots."""

    def __init__(self, *args, **kwargs):
        """Convert ``*args`` and ``**kwargs`` into dot-attribute dict.

        :param *args: varags
        :param **kwargs: varags
        """
        super().__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

    def __getattr__(self, attr):
        """doctstring."""
        return self.get(attr)

    def __setattr__(self, key, value):
        """doctstring."""
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        """doctstring."""
        super().__setitem__(key, value)
        if isinstance(value, dict):
            value = Config(**value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        """doctstring."""
        self.__delitem__(item)

    def __delitem__(self, key):
        """doctstring."""
        super().__delitem__(key)
        del self.__dict__[key]



"""TODO docstring."""
from pathlib import Path
from typing import List, Union, Dict, Optional
from uuid import UUID, uuid4
import re

import pandas as pd
from pydantic import (
    BaseModel,
    Field,
    AnyHttpUrl,
    validator,
)

from seminar.models import helpers


class _OptionalURLs(BaseModel):
    slides: Optional[str] = ""
    video: Optional[str] = ""

    class Config:
        alias_generator = helpers.dashify

    @validator("slides")
    def validate_slides(cls, url: str) -> str:
        if "docs" in url and "presentation" in url:
            docs_base_url = "(?:https?://)?docs.google.com/presentation/d/e/"

            url = re.sub(f"{docs_base_url}|$", "", url)
            url = url.split("/", maxsplit=1)[0]

        test_urls = [
            "https://docs.google.com/presentation/d/e/{{ slug }}/embed",
            "https://ucfai.org/slides/{{ slug }}",
        ]

        return helpers.test_url(url, test_urls)

    @validator("video")
    def validate_videos(cls, url: str) -> str:
        if "youtu" in url:
            scheme = "(:?https?://)?"
            ln_old = "(?:www\.)?youtube.com/watch?v="
            ln_new = "youtu.be/"
            yt_full = f"{scheme}(?:{ln_old}|{ln_new})"
            url = re.sub(f"{yt_full}|$", "", url)
            url = re.search("([A-Za-z0-9-_]{9,11})", url).group(0)

        test_urls = [
            "https://youtube.com/embed/{{ slug }}",
        ]

        return helpers.test_url(url, test_urls)


class _OptionalKaggle(BaseModel):
    kernels: Union[List[str], bool] = []
    datasets: Union[List[str], bool] = []
    competitions: Union[List[str], bool] = []
    enable_gpu: bool = False

    class Config:
        alias_generator = helpers.dashify

    @validator("kernels", "datasets", each_item=True)
    def is_kaggle_slug_valid(cls, slug: str) -> str:
        try:
            assert slug.count("/") == 1
        except AssertionError:
            raise ValueError(f"Invalid slug: {slug}. Check the Kaggle API.")
        finally:
            return slug

    @validator("kernels", "datasets", "competitions", each_item=True)
    def is_kaggle_slug_live(cls, slug):
        test_urls = ["https://kaggle.com", "https://kaggle.com/c"]
        # kernel-metadata.json only accepts slugs. No error means it's live.
        _ = helpers.test_url(slug, test_urls)
        return slug


class Meeting(BaseModel):
    id: Union[str, UUID] = Field(default_factory=uuid4)
    title: str
    filename: str
    date: pd.Timestamp
    authors: List[str] = []
    cover_image: Union[str, AnyHttpUrl] = ""
    tags: List = []
    room: str = ""
    abstract: str = ""

    urls: Optional[_OptionalURLs] = {}
    kaggle: Optional[Union[_OptionalKaggle, bool]] = {}
    papers: Optional[Dict[str, AnyHttpUrl]] = {}

    class Config:
        alias_generator = helpers.dashify

    _normalize_filename = validator("filename")(helpers.downcase_dashify_name)

    @validator("papers")
    def normalize_paperkey(cls, papers) -> Dict:
        if papers is None:
            return None

        assert type(papers) == dict

        cleaned = {}
        for key, url in papers.items():
            normed = helpers.downcase_dashify_name(key)
            url = helpers.test_url(url)
            cleaned[normed] = url

        return cleaned

    def __repr__(self):
        s = self.filename

        if not pd.isnull(self.date):
            return f"{self.date.isoformat()[5:10]}-{s}"

        return s

    def __str__(self):
        return repr(self)

    def as_dir(self, prefix: Path = None):
        """Returns a ``Path`` of a ``Meeting`` as a directory.

        (normally used to place contents, e.g. Notebooks, Markdown, data, PDFs, etc.)
        """
        folder = Path(repr(self))
        if prefix:
            return prefix / folder
        return folder

    def as_file(self, prefix: Path = None, suffix: str = ""):
        """Returns ``Path`` of a ``Meeting`` as a file.

        (normally used for Notebook and Markdown files)
        """
        file: Path = self.as_dir(prefix) / self.filename
        if suffix:
            return file.with_suffix(suffix)
        return file

    @property
    def path(self) -> Path:
        return Path(repr(self))

    @property
    def file(self) -> Path:
        return self.path / self.filename

    @property
    def markdown(self) -> Path:
        return self.file.with_suffix(".md")

    @property
    def solnbook(self) -> Path:
        return self.file.with_suffix(".solution.ipynb")

    @property
    def workbook(self) -> Path:
        return self.file.with_suffix(".ipynb")

    @property
    def kernel(self) -> Path:
        return self.file.with_name("kernel-metadata.json")




"""TODO docstring
"""
from typing import Union

from click import Context

from seminar.utils import Config
from seminar.models import Meeting


def for_meeting(cfg: Config, value: Union[str, Meeting]) -> Meeting:
    """Searches for the nearest matching meeting.

    :param cfg: Config
    :param query: Union[str, Meeting]
    """
    # NOTE solution at https://stackoverflow.com/a/36438326/2714651
    if not value:
        cfg.meetings = cfg.syllabus
    elif isinstance(value, Meeting):
        cfg.meetings = [value]
    elif type(value) == str:

        def is_substr(m: Meeting):
            return value.lower() in m.filename.lower()

        try:
            bytitle = filter(is_substr, cfg.syllabus)
            cfg.meetings = [next(bytitle)]
        except StopIteration:
            mtg_ls = "\n- ".join(cfg.syllabus)
            raise ValueError(
                f"Failed to find a `Meeting` containing `{value}`. "
                f"Here's a current list of meetings:{mtg_ls}"
            )
    else:
        raise ValueError("`query` must be a `str` or `Meeting`.")

    cfg.query = cfg.meetings[0]

    return value



from typing import Union

from seminar.models import Meeting


def search(ctx, query: Union[str, Meeting] = None) -> Union[list, Meeting]:
    # NOTE solution at https://stackoverflow.com/a/36438326/2714651
    # @functools.wraps(func)
    # def wrapper(ctx, *args, **kwargs):
    if not query:
        return ctx.syllabus
    elif isinstance(query, Meeting):
        return query
    elif type(query) == str:
        is_substr = lambda x: query.lower() in x.filename.lower()
        bytitle = filter(is_substr, ctx.syllabus)
        try:
            return next(bytitle)
        except StopIteration:
            mtg_ls = "\n- ".join(ctx.syllabus)
            raise ValueError(
                f"Failed to find a `Meeting` containing `{query}`. "
                f"Here's a current list of meetings:{mtg_ls}"
            )
    else:
        raise ValueError("`query` must be of type: {`str`, `Meeting`}.")
